@@ -27,6 +27,7 @@
 ########################################################################
 '''implementation of the ``InteractiveClean`` application for interactive control
 of tclean'''
+import os
 import copy
 import asyncio
 import shutil
@@ -48,6 +49,12 @@ from casagui.utils import find_ws_address, convert_masks
 from casagui.toolbox import CubeMask
 from casagui.bokeh.components import SVGIcon
 from casagui.bokeh.sources import DataPipe
+from ..utils import DocEnum
+
+class MaskMode(DocEnum):
+    '''Different masking modes available in addition to a user supplied mask'''
+    PB = 1, 'primary beam mask'
+    AUTOMT = 2, 'multi-threshold auto masking'
 
 class InteractiveClean:
     '''InteractiveClean(...) implements interactive clean using Bokeh
@@ -101,6 +108,8 @@ class InteractiveClean:
             be specified, e.g. "ngc5921.ms", or a list of paths may be specified, e.g.
             ``['ngc5921a.ms','ngc5921b.ms']``.
         imagename (str): Name of the output images (without a suffix).
+        mask (str or MaskMode): user specified CASA imaging mask cube (path) or MaskMode enum indicating
+            mask to generate
         imsize (int, :obj:`list` of int): Number of pixels in a image plane, for example ``[350,250]``.
             ``500`` implies ``[500,500]``. The number of pixels must be even and factorizable by 2,3,5,7 only
             due to FFT limitations.
@@ -142,9 +151,17 @@ class InteractiveClean:
         self._error_result = err
         self.__stop( )
 
-    def __init__( self, vis, imagename, field='', spw='', imsize=[100], cell=[ ], phasecenter='', stokes='I', specmode='cube', nchan=-1, start='',
-                  width='', interpolation='linear', gridder='standard', pblimit=0.2, deconvolver='hogbom', niter=0,
-                  threshold='0.1Jy', cycleniter=-1, cyclefactor=1.0, scales=[], weighting='natural', robust=float(0.5), nmajor=1 ):
+    def __init__( self, vis, imagename, mask=None, field='', spw='', timerange='', uvrange='', antenna='', scan='', intent='', datacolumn='corrected',
+                  nterms=int(2), imsize=[100], cell=[ ], phasecenter='', stokes='I', specmode='cube', reffreq='', nchan=-1, start='', width='',
+                  outframe='LSRK', interpolation='linear', perchanweightdensity=True, gridder='standard', wprojplanes=int(1), mosweight=True,
+                  psterm=False, wbawp=True, usepointing=False, conjbeams=False, pointingoffsetsigdev=[  ], pblimit=0.2, deconvolver='hogbom',
+                  niter=0, threshold='0.1Jy', nsigma=0.0, cycleniter=-1, cyclefactor=1.0, scales=[], restoringbeam='', pbcor=False,
+                  weighting='natural', robust=float(0.5), npixels=0, gain=float(0.1), sidelobethreshold=3.0, noisethreshold=5.0,
+                  lownoisethreshold=1.5, negativethreshold=0.0, minbeamfrac=0.3, growiterations=75, dogrowprune=True, minpercentchange=-1.0,
+                  fastnoise=True, savemodel='none', parallel=False, nmajor=1 ):
+
+        if deconvolver == 'mtmfs':
+            raise RuntimeError("deconvolver task does not support 'mtmf' deconvolver")
 
         ###
         ### used by data pipe (websocket) initialization function
@@ -158,18 +175,40 @@ class InteractiveClean:
         self._color = { 'residual': 'black',
                         'flux':     'forestgreen' }
 
+        ###
+        ### Auto Masking et al.
+        ###     if the user has specified a mask cube, it OVERRIDES the default, generated mask name
+        self._mask_path = ''
+        self._usemask = 'user'
+        if isinstance( mask, MaskMode ):
+            if mask == MaskMode.AUTOMT:
+                self._usemask = 'auto-multithresh'
+            elif mask == MaskMode.PB:
+                self._usemask = 'pb'
+            else:
+                raise RuntimeError( 'internal consistence error for MaskMode' )
+        elif isinstance( mask, str ) and os.path.isdir( mask ):
+            ### override default mask name
+            self._mask_path = mask
+
         ###
         ### clean generator
         ###
         if _gclean is None:
             raise RuntimeError('casatasks gclean interface is not available')
 
-        self._clean = _gclean( vis=vis, imagename=imagename, field=field, spw=spw, imsize=imsize, cell=cell, phasecenter=phasecenter,
-                               stokes=stokes, specmode=specmode, nchan=nchan, start=start, width=width, interpolation=interpolation,
-                               gridder=gridder, pblimit=pblimit, deconvolver=deconvolver, niter=niter, threshold=threshold,
-                               cycleniter=cycleniter, cyclefactor=cyclefactor, scales=scales, weighting=weighting, robust=robust, nmajor=nmajor,
-                               history_filter=lambda index, arg, history_value: ( f'mask=masks[{len(self._mask_history)-1}]' if len(self._mask_history) > 0 else '' ) \
-                                                                                  if arg == 'mask' else history_value )
+        self._clean = _gclean( vis=vis, imagename=imagename, field=field, spw=spw, timerange=timerange,  uvrange=uvrange, antenna=antenna, scan=scan,
+                               intent=intent, datacolumn=datacolumn, nterms=nterms, imsize=imsize, cell=cell, phasecenter=phasecenter, stokes=stokes,
+                               specmode=specmode, reffreq=reffreq, nchan=nchan, start=start, width=width, outframe=outframe, interpolation=interpolation,
+                               perchanweightdensity=perchanweightdensity, gridder=gridder, wprojplanes=wprojplanes, mosweight=mosweight, psterm=psterm,
+                               wbawp=wbawp, usepointing=usepointing, conjbeams=conjbeams, pointingoffsetsigdev=pointingoffsetsigdev, pblimit=pblimit,
+                               deconvolver=deconvolver, niter=niter, threshold=threshold, nsigma=nsigma, cycleniter=cycleniter, cyclefactor=cyclefactor,
+                               scales=scales, restoringbeam=restoringbeam, pbcor=pbcor, weighting=weighting, robust=robust, npixels=npixels, gain=gain,
+                               sidelobethreshold=sidelobethreshold, noisethreshold=noisethreshold, lownoisethreshold=lownoisethreshold,
+                               negativethreshold=negativethreshold, minbeamfrac=minbeamfrac, growiterations=growiterations, dogrowprune=dogrowprune,
+                               minpercentchange=minpercentchange, fastnoise=fastnoise, savemodel=savemodel, parallel=parallel, nmajor=nmajor,
+                               usemask=self._usemask, mask=self._mask_path
+                      )
         ###
         ### self._convergence_data: accumulated, pre-channel convergence information
         ###                         used by ColumnDataSource
@@ -183,7 +222,8 @@ class InteractiveClean:
         self._convergence_id = str(uuid4( ))
         #print(f'convergence:',self._convergence_id)
 
-        self._status['log'] = Div( text='''<hr style="width:790px"><p style="width:790px">%s</p>''' % self._clean.cmds( )[-1] )
+        self._status['log'] = Div( text='''<hr style="width:790px">%s''' % ''.join([ f'<p style="width:790px">{cmd}</p>' for cmd in self._clean.cmds( )[-2:] ]) )
+        ###                        >>>--------tclean+deconvolve----------------------------------------------------------------------------------------^^^^^
         self._status['stopcode'] = Div( text="<div>initial image</div>" )
 
         ###
@@ -204,6 +244,8 @@ class InteractiveClean:
         ### GUI Elements
         self._imagename = imagename
         self._residual_path = ("%s.residual" % imagename) if self._clean.has_next() else (self._clean.finalize()['image'])
+        if not os.path.isdir(self._mask_path):
+            self._mask_path = ("%s.mask" % imagename) if self._clean.has_next() else None
         self._pipe = { 'control': None, 'converge': None }
         self._control = { }
         self._cb = { }
@@ -356,6 +398,10 @@ class InteractiveClean:
                                            }''',
 
                        'clean-gui-update': '''function update_gui( msg ) {
+                                               if ( msg.result === 'no-action' ) {
+                                                   update_status( 'nothing done' )
+                                                   enable( false )
+                                               } else
                                                if ( msg.result === 'update' ) {
                                                    if ( 'cmd' in msg ) {
                                                        log.text = log.text + msg.cmd
@@ -363,7 +409,11 @@ class InteractiveClean:
                                                    refresh( msg )
                                                    // stopcode == 1: iteration limit hit
                                                    // stopcode == 9: major cycle limit hit
-                                                   state.stopped = state.stopped || (msg.stopcode > 1 && msg.stopcode < 9) || msg.stopcode == 0
+                                                   // *******************************************************************************************
+                                                   // ******** perhaps the user should not be locked into exiting after the limit is hit ********
+                                                   // *******************************************************************************************
+                                                   //state.stopped = state.stopped || (msg.stopcode > 1 && msg.stopcode < 9) || msg.stopcode == 0
+                                                   state.stopped = false
                                                    if ( state.mode === 'interactive' && ! state.awaiting_stop ) {
                                                        btns['stop'].button_type = "danger"
                                                        update_status( 'stopcode' in msg ? msg.stopcode : -1 )
@@ -411,8 +461,7 @@ class InteractiveClean:
                                            }''',
                    }
 
-
-        self._cube = CubeMask( self._residual_path, abort=self._abort_handler )
+        self._cube = CubeMask( self._residual_path, mask=self._mask_path, abort=self._abort_handler )
         ###
         ### error or exception result
         ###
@@ -461,21 +510,27 @@ class InteractiveClean:
                     msg['value']['mask'] = ''
                 self._clean.update( { **msg['value'] } )
                 stopcode, self._convergence_data = await self._clean.__anext__( )
-                if stopcode > 1 and stopcode < 9: # 1: iteration limit hit, 9: major cycle limit hit
-                    self._clean.finalize()
+                # *******************************************************************************************
+                # ******** perhaps the user should not be locked into exiting after the limit is hit ********
+                # *******************************************************************************************
+                #if stopcode > 1 and stopcode < 9: # 1: iteration limit hit, 9: major cycle limit hit
+                #    self._clean.finalize()
 
                     # self._cube.update_image(self._clean.finalize()['image']) # TODO show the restored image
                 if len(self._convergence_data) == 0 and stopcode == 7:
                     return dict( result='error', stopcode=stopcode, cmd=f"<p>mask error encountered (stopcode {stopcode})</p>", convergence=None  )
+                if len(self._convergence_data) == 0:
+                    return dict( result='no-action', stopcode= stopcode, cmd=f'<p style="width:790px">no operation</p>',
+                                 convergence=None, iterdone=0 )
                 if len(self._convergence_data) * len(self._convergence_data[0]) > self._threshold_chan or \
                    len(self._convergence_data[0][0]['iterations']) > self._threshold_iterations:
-                    return dict( result='update', stopcode=stopcode, cmd=f'<p style="width:790px">{self._clean.cmds( )[-1]}</p>',
+                    return dict( result='update', stopcode=stopcode, cmd=''.join([ f'<p style="width:790px">{cmd}</p>' for cmd in self._clean.cmds( )[-2:] ]),
                                  convergence=None, iterdone=sum([ x['iterations'][1]  for y in self._convergence_data.values() for x in y.values( ) ]) )
                 else:
-                    return dict( result='update', stopcode=stopcode, cmd=f'<p style="width:790px">{self._clean.cmds( )[-1]}</p>',
+                    return dict( result='update', stopcode=stopcode, cmd=''.join([ f'<p style="width:790px">{cmd}</p>' for cmd in self._clean.cmds( )[-2:] ]),
                                  convergence=self._convergence_data, iterdone=sum([ x['iterations'][1]  for y in self._convergence_data.values() for x in y.values( ) ]) )
 
-                return dict( result='update', stopcode=stopcode, cmd=f'<p style="width:790px">{self._clean.cmds( )[-1]}</p>',
+                return dict( result='update', stopcode=stopcode, cmd=''.join([ f'<p style="width:790px">{cmd}</p>' for cmd in self._clean.cmds( )[-2:] ]),
                              convergence=self._convergence_data, iterdone=sum([ x['iterations'][1]  for y in self._convergence_data.values() for x in y.values( ) ]) )
             elif msg['action'] == 'stop':
                 self.__stop( )
@@ -674,6 +729,9 @@ class InteractiveClean:
         # spectra
         # convergence
         # log
+
+        mask_color_pick, mask_alpha_pick, mask_clean_notclean_pick = self._cube.bitmask_controls( )
+
         self._fig['layout'] = column(
                                   row(
                                       column(
@@ -692,6 +750,10 @@ class InteractiveClean:
                                       ),
                                       column( self._fig['image'],
                                               row( Spacer(height=self._control['help'].height, sizing_mode="scale_width"),
+                                                   self._cube.palette( ),
+                                                   mask_clean_notclean_pick,
+                                                   mask_color_pick,
+                                                   mask_alpha_pick,
                                                    self._control['help'],
                                                    Spacer(height=self._control['help'].height, width=30)
                                               )
